/*
	SambaScanner - A tool to scan Samba networks for files.
	Copyright (C) 2005 Johannes Bauer
	
	This file is part of SambaScanner.

    SambaScanner is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; this program is ONLY licensed under
	version 2 of the License, later versions are explicitly excluded.

    SambaScanner is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with SambaScanner; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA

	Johannes Bauer
	Mussinanstr. 140
	92318 Neumarkt/Opf.
	JohannesBauer@gmx.de
*/

#include <stdio.h>
#include <sys/types.h>
#include <unistd.h>
#include <stdlib.h>
#include <errno.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <signal.h>
#include <sys/wait.h>
#include <time.h>
#include <pthread.h>
#include <sys/timeb.h>
#include <getopt.h>

#include "Mutex.h"
#include "Parser.h"

#define VERSION				"0.07"
#define FILE_WORKGROUPINFO	"/tmp/sambascanner_workgroup.tmp"
#define FILE_MASTERBROWSER	"/tmp/sambascanner_masterbrowser.tmp"

#define message(msglevel, args...) if (msglevel<=Options.DebugLvl) fprintf(stderr, args)

#ifndef DEFAULT_SMBCLIENT
#define DEFAULT_SMBCLIENT	"/usr/bin/smbclient"
#endif

#ifndef DEFAULT_NMBLOOKUP
#define DEFAULT_NMBLOOKUP	"/usr/bin/nmblookup"
#endif

#ifndef DEFAULT_SMBLISTER
#define DEFAULT_SMBLISTER	"/usr/bin/smblister"
#endif


struct Process {
	pid_t PID;
	int Running;
	int Status;
	int Signal;			/* Was killed by signal? */
	int Valid;			/* Is the entered PID already valid? */
	double TermTime;
	double KillTime;
};

struct Options_Structure {
	double Timings;
	char Workgroup[256];
	int DebugLvl;
	char NmbLookup_Executable[256];
	char SmbClient_Executable[256];
	char SmbLister_Executable[256];
	char Outfile[256];
} Options;

struct Process *Processes = NULL;
int NumberProcesses = 0;

int NumberHosts = 0;
char **Hostlist = NULL;
int NumberShares = 0;
char **Sharelist = NULL;
int NumberShareResultFiles = 0;
char **Shareresults = NULL;

int Hostlookup_Processes_Succeeded = 0;
int Hostlookup_Processes_Failed = 0;

double ftm() {
	struct timeb TimeB;
	ftime(&TimeB);
	return ((double)TimeB.time+((double)TimeB.millitm/1000.0));
}

void SIGCHLD_Handler(int Signal) {
	int i;
	pid_t PID;
	for (i=0; i<NumberProcesses; i++) {
		if (Processes[i].Valid==1) {
			PID = waitpid(Processes[i].PID, &Processes[i].Status, WNOHANG);
/*			printf("Handling SIGCHLD [%d] => PID = %d, result = %d...\n", i, Processes[i].PID, PID); */
			if (PID>0) {
				/* Child really exited */
				Processes[i].Running = 0;
			}
		} else {
/*			printf("Invalid process %d ignored!\n", i); */
		}
	}
}

void Install_SigHandler() {
	struct sigaction Action;
	sigset_t Mask;
	
	sigemptyset(&Mask);
	Action.sa_handler = SIGCHLD_Handler;
	Action.sa_flags = SA_RESTART;
	Action.sa_mask = Mask;
	
	sigaction(SIGCHLD, &Action, NULL);
}

int Exec_Timeout(const char *Command, const char *Parameter, double TermTime, double KillTime, char *Outfile) {
	pid_t PID;
	char **Arguments = NULL;
	int ArgumentCount = 0;
	char *Token;
	char *String;
	double Time;
	int CurrentProcess;
	sigset_t Mask;
	
	Lock_Exec();
	Time=ftm();
	Processes = (struct Process*)realloc(Processes, sizeof(struct Process)*(NumberProcesses+1));
	Processes[NumberProcesses].Running = 1;
	Processes[NumberProcesses].Status = 0;
	Processes[NumberProcesses].Signal = 0;
	Processes[NumberProcesses].Valid = 0;
	Processes[NumberProcesses].TermTime = Time+TermTime;
	Processes[NumberProcesses].KillTime = Time+TermTime+KillTime;
	NumberProcesses++;
	CurrentProcess=NumberProcesses-1;
	Unlock_Exec();	

	sigemptyset(&Mask);
	sigaddset(&Mask, SIGCHLD);
	sigprocmask(SIG_BLOCK, &Mask, NULL);
	
	PID = fork();
	if (PID==0) {
		/* Child process, parse arguments */
		String = (char*)malloc(strlen(Parameter)+1);
		if (String==NULL) {
			message(1, "Malloc failed when allocating string: %s\n", strerror(errno));
			exit(1);
		}

		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = (char*)malloc((strlen(Command)*sizeof(char))+1);
		strcpy(Arguments[ArgumentCount-1], Command);
		
		strcpy(String, Parameter);
		Token = strtok(String, " ");
		while (Token!=NULL) {
			ArgumentCount++;
			Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
			Arguments[ArgumentCount-1] = (char*)malloc((strlen(Token)*sizeof(char))+1);
			strcpy(Arguments[ArgumentCount-1], Token);
			Token = strtok(NULL, " ");
		}
		ArgumentCount++;
		Arguments = (char**)realloc(Arguments, sizeof(char**)*ArgumentCount);
		Arguments[ArgumentCount-1] = NULL;

		{
			int fd;
			int Old_Stderr;
			fd = open(Outfile, O_CREAT | O_TRUNC | O_WRONLY, 0644);
			if (fd==-1) {
				message(1, "File open error: %s\n", strerror(errno));
				exit(1);
			}
			Old_Stderr=dup(STDERR_FILENO);
			dup2(fd, STDOUT_FILENO);
			dup2(fd, STDERR_FILENO);
			execv(Command, Arguments);
			dup2(Old_Stderr, STDERR_FILENO);
			message(1, "Execution of '%s' failed: %s\n", Command, strerror(errno));
		}
		exit(1);
	} else {
		/* Parent process */
		Processes[CurrentProcess].PID = PID;
		Processes[CurrentProcess].Valid = 1;
		message(2, "Forked off child '%s %s' with PID %d.\n", Command, Parameter, PID);
	}
	
	sigprocmask(SIG_UNBLOCK, &Mask, NULL);
	return CurrentProcess;
}

int WaitFor(int ProcID) {
	double Before, After;
	int Return;
	
	Before=ftm();
	while (Processes[ProcID].Running==1) {
/*		printf("Process %d state: pid=%d running=%d return=%d\n", ProcID, Processes[ProcID].PID, Processes[ProcID].Running, Processes[ProcID].Status); */
		usleep(100*1000);	/* 100ms */
	}
	After=ftm();
	if (Processes[ProcID].Signal==0) Return=WEXITSTATUS(Processes[ProcID].Status);
		else Return=-Processes[ProcID].Signal;
	message(2, "Waited for thread %d %.2f seconds, returning with %d.\n", ProcID, After-Before, Return);
	return Return;
}

void* Process_Watcher(void *Argument) {
	int i;
	double Time;
	
	while (1) {
		Time=ftm();
		for (i=0; i<NumberProcesses; i++) {
			if ((Processes[i].Running==1) && (Processes[i].Valid==1)) {
				if (Time>(Processes[i].KillTime*Options.Timings)) {
					message(3, "=> SIGKILL %d [PID %d]\n", i, Processes[i].PID);
					Processes[i].Signal = SIGKILL;
					kill(Processes[i].PID, SIGKILL);
				} else {
					if (Time>(Processes[i].TermTime*Options.Timings)) {
						message(3, "=> SIGTERM %d [PID %d]\n", i, Processes[i].PID);
						Processes[i].Signal = SIGTERM;
						kill(Processes[i].PID, SIGTERM);
					}
				}
			}
		}
		usleep(100*1000);	/* 100ms */
	}
}

void* Process_Host(void *Void_Hostname) {
	char *Hostname = (char*)Void_Hostname;
	char Buffer[256];
	char FileBuffer[256];
	int Result, ProcID;

	strcpy(FileBuffer, "/tmp/process_hostsXXXXXX");
	mkstemp(FileBuffer);

	message(2, "Processing host: '%s'\n", Hostname);
	snprintf(Buffer, 255, "-g -N -L //%s", Hostname);
	ProcID = Exec_Timeout(Options.SmbClient_Executable, Buffer, 2.5, 1.0, FileBuffer);
	Result = WaitFor(ProcID);
	if (Result<0) {
		/* Process was killed */
		Hostlookup_Processes_Failed++;
	} else {
		Hostlookup_Processes_Succeeded++;
	}
	Parse_Hostinfo(Hostname, FileBuffer, &NumberHosts, &Hostlist, &NumberShares, &Sharelist);
	unlink(FileBuffer);
	HostCnt_Dec();
	return NULL;
}

void* Process_Share(void *Void_Sharename) {
	char *Sharename = (char*)Void_Sharename;
	char Buffer[256];
	char FileBuffer[256];
	int Result, ProcID;

	strcpy(FileBuffer, "/tmp/process_sharesXXXXXX");
	mkstemp(FileBuffer);
	Add_String(FileBuffer, &NumberShareResultFiles, &Shareresults);
	
	message(2, "Now revealing content of smb:%s\n", Sharename);
	snprintf(Buffer, 255, "smb:%s", Sharename);
	ProcID = Exec_Timeout(Options.SmbLister_Executable, Buffer, 90.0, 5.0, FileBuffer);
	Result = WaitFor(ProcID);
	
	HostCnt_Dec();
	return NULL;
}

void Display_Help(char **argv) {
	fprintf(stderr, "Syntax: %s [-t Time] [-w Workgroup] [-d Debuglevel]\n", argv[0]);
	fprintf(stderr, "\n");
	fprintf(stderr, "-t, --time       Requires a float argument which states by which factor the\n");
	fprintf(stderr, "                 internal timeouts should be multiplied with.\n");
	fprintf(stderr, "-w, --workgroup  Manually sets the workgroup to the given argument\n");
	fprintf(stderr, "-d, --debug      Set the debug level from 0 (silent) to 3 (full debug),\n");
	fprintf(stderr, "                 default level is 1\n");
	fprintf(stderr, "-s, --smbclient  Set the path where smbclient can be found (default " DEFAULT_SMBCLIENT ").\n");
	fprintf(stderr, "-n, --nmblookup  Set the path where nmblookup can be found (default " DEFAULT_NMBLOOKUP ").\n");
	fprintf(stderr, "-l, --smblister  Set the path where smblister can be found (default " DEFAULT_SMBLISTER ").\n");
	fprintf(stderr, "-c, --computer   Specify a computer on the network.\n");
	fprintf(stderr, "-o, --outfile    Set the output file (default 'Filelist.txt')\n");
	fprintf(stderr, "-h, --help       Display this help\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "Example: %s -w MYGROUP -t 2.5\n", argv[0]);
	fprintf(stderr, "         Multiplies all timings by 2.5 and sets the workgroup to 'MYGROUP'.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "         %s -w MATRIX -c NEO\n", argv[0]);
	fprintf(stderr, "         Select workgroup 'MYGROUP' and specify known computer 'NEO'.\n");
	fprintf(stderr, "\n");
	fprintf(stderr, "This is SambaScanner " VERSION ", released under the GNU GPL-2.\n");
}

int Process_Parameters(int argc, char **argv) {
	Options.Workgroup[0]=0;
	Options.Timings=1.0;
	Options.DebugLvl=1;
	strcpy(Options.SmbClient_Executable, DEFAULT_SMBCLIENT);
	strcpy(Options.NmbLookup_Executable, DEFAULT_NMBLOOKUP);
	strcpy(Options.SmbLister_Executable, DEFAULT_SMBLISTER);
	strcpy(Options.Outfile, "Filelist.txt");

	while (1) {
		int Result;
		int Option_Index = 0;
		static struct option Long_Options[] = {
			{"time", 1, 0, 0},
			{"workgroup", 1, 0, 0},
			{"help", 0, 0, 0},
			{"debug", 1, 0, 0},
			{"smbclient", 1, 0, 0},
			{"nmblookup", 1, 0, 0},
			{"smblister", 1, 0, 0},
			{"outfile", 1, 0, 0},
			{"computer", 1, 0, 0},
			{0, 0, 0, 0}
		};
		Result = getopt_long (argc, argv, "t:w:hd:s:n:l:o:c:", Long_Options, &Option_Index);
		if (Result==-1) break;

		switch (Result) {
			case 0: {
				switch (Option_Index) {
					case 0: { /* --time */
						double Timeout;
						Timeout=atof(optarg);
						if ((Timeout<0.1) || (Timeout>1000.0)) {
							fprintf(stderr, "Error: Timeout must be inbetween 0.1 and 1000\n");
							return 0;
						}
						Options.Timings=Timeout;
						break;
					}
					case 1: { /* --workgroup */
						strncpy(Options.Workgroup, optarg, 255);
						break;
					}
					case 2: { /* --help */
						Display_Help(argv);
						return 0;
						break;
					}
					case 3: { /* --debug */
						int DebugLvl;
						DebugLvl=atoi(optarg);
						if ((DebugLvl<0) || (DebugLvl>3)) {
							fprintf(stderr, "Error: Debuglevel must be inbetween 0 and 3\n");
							return 0;
						}
						Options.DebugLvl=DebugLvl;
						break;
					}
					case 4: { /* --smbclient */
						strncpy(Options.SmbClient_Executable, optarg, 255);
						break;
					}
					case 5: { /* --nmblookup */
						strncpy(Options.NmbLookup_Executable, optarg, 255);
						break;
					}
					case 6: { /* --smblister */
						strncpy(Options.SmbLister_Executable, optarg, 255);
						break;
					}
					case 7: { /* --outfile */
						strncpy(Options.Outfile, optarg, 255);
						break;
					}
					case 8: { /* --computer */
						Add_String(optarg, &NumberHosts, &Hostlist);
						break;
					}
				}
				break;
			}

			case 't': {
				double Timeout;
				Timeout=atof(optarg);
				if ((Timeout<0.1) || (Timeout>1000.0)) {
					fprintf(stderr, "Error: Timeout must be inbetween 0.1 and 1000\n");
					return 0;
				}
				Options.Timings=Timeout;
				break;
			}
					  
			case 'w': {
				strncpy(Options.Workgroup, optarg, 255);
				break;
			}

			case 'd': {
				int DebugLvl;
				DebugLvl=atoi(optarg);
				if ((DebugLvl<0) || (DebugLvl>3)) {
					fprintf(stderr, "Error: Debuglevel must be inbetween 0 and 3\n");
					return 0;
				}
				Options.DebugLvl=DebugLvl;
				break;
			}
			
			case 's': {
				strncpy(Options.SmbClient_Executable, optarg, 255);
				break;
			}
			case 'n': {
				strncpy(Options.NmbLookup_Executable, optarg, 255);
				break;
			}
			case 'l': {
				strncpy(Options.SmbLister_Executable, optarg, 255);
				break;
			}
			case 'o': {
				strncpy(Options.Outfile, optarg, 255);
				break;
			}
			case 'c': { /* Computer */
				Add_String(optarg, &NumberHosts, &Hostlist);
				break;
			}
					  
			case '?':
			case 'h': {
				Display_Help(argv);
				return 0;
				break;
			}

			default: {
				fprintf(stderr, "Returned character code was 0x%x\n", Result);
				Display_Help(argv);
				return 0;
				break;
			}
		}
	}

	if (optind<argc) {
		fprintf(stderr, "Invalid argument: ");
		while (optind<argc) fprintf(stderr, "%s ", argv[optind++]);
		fprintf(stderr, "\n");
		Display_Help(argv);
		return 0;
	}

	return 1;
}

char *SingularPlural(char *Singular, char *Plural, int Number) {
	return (Number==1) ? Singular : Plural;
}

int main(int argc, char **argv) {
	int ProcID;
	int Result;
	pthread_t ProcessWatcher;
	int i;
	pthread_t TempThread;
	char Buffer[256];
		
	Init_Semaphores();
	if (!Process_Parameters(argc, argv)) {
		return 1;
	};
	Install_SigHandler();
	pthread_create(&ProcessWatcher, NULL, Process_Watcher, NULL);
	
	if (Options.Workgroup[0]==0) {
		/* Workgroup was not set in options */
		message(1, "Looking up workgroup...\n");
		ProcID = Exec_Timeout(Options.NmbLookup_Executable, "-M -S -- -", 10.0, 3.0, FILE_WORKGROUPINFO);
		Result = WaitFor(ProcID);
		if (Result!=0) {
			message(1, "Process terminated at looking up workgroup.\n");
			return 1;
		}
		if (Parse_Groupinfo(FILE_WORKGROUPINFO, Options.Workgroup)!=0) {
			message(1, "Process terminated at parsing workgroup info list.\n");
			return 2;
		};
		message(2, "Returned workgroup: '%s'\n", Options.Workgroup);
		unlink(FILE_WORKGROUPINFO);
	
		if (strlen(Options.Workgroup)==0) {
			fprintf(stderr, "No workgroup found, aborting!\n");
			fprintf(stderr, "Try setting the workgroup manually.\n");
			return 3;
		}
	}
	
	snprintf(Buffer, 255, "-M -S %s", Options.Workgroup);
	message(1, "Retrieving master browsers in workgroup '%s'...\n", Options.Workgroup);
	ProcID = Exec_Timeout(Options.NmbLookup_Executable, Buffer, 10.0, 3.0, FILE_MASTERBROWSER);
	Result = WaitFor(ProcID);
	if (Result!=0) {
		message(1, "Process terminated at retrieving master browsers.\n");
		return 1;
	}
	if (Parse_Masterbrowser(FILE_MASTERBROWSER, &NumberHosts, &Hostlist)!=0) {
		message(1, "Process terminated at parsing master browser list.\n");
		return 1;
	}
	unlink(FILE_MASTERBROWSER);
	
	{
		int NumberKilled, NumberOkay;
		NumberKilled=0;
		NumberOkay=0;
		for (i=0; i<NumberHosts; i++) {
			HostCnt_Inc();
			pthread_create(&TempThread, NULL, Process_Host, Hostlist[i]);
			if (i==NumberHosts-1) {
				Waitfor_HostCnt();
			}
		}
		message(2, "%d processes killed (%.1f%%), %d returned sucessfully (%.1f%%).\n", 
				Hostlookup_Processes_Failed, 
				100.0*Hostlookup_Processes_Failed/(Hostlookup_Processes_Failed+Hostlookup_Processes_Succeeded), 
				Hostlookup_Processes_Succeeded, 
				100.0*Hostlookup_Processes_Succeeded/(Hostlookup_Processes_Failed+Hostlookup_Processes_Succeeded)
			);
		message(2, "Total hosts found: %d\n",NumberHosts);
	}

	/*
	{
		int i;
		fprintf(stderr, "Number of hosts in list: %d\n", NumberHosts);
		for (i=0; i<NumberHosts; i++) {
			fprintf(stderr, "%3d: %s\n", i+1, Hostlist[i]);
		}
		fprintf(stderr, "\n");
	}
	*/

	/* Now try to reveal all share content on the found shares */
	{
		int i;
		message(1, "Revealing share content of %d %s in %d %s...\n", NumberShares, SingularPlural("share", "shares", NumberShares), NumberHosts, SingularPlural("host", "hosts", NumberHosts));
		for (i=0; i<NumberShares; i++) {
			HostCnt_Inc();
			pthread_create(&TempThread, NULL, Process_Share, Sharelist[i]);
			if (i==NumberHosts-1) {
				Waitfor_HostCnt();
			}
		}
	}

	{
		int i;
		FILE *f, *g;
		char Line[1024];
		f=fopen(Options.Outfile, "w");
		if (f==NULL) {
			message(1, "%s: Error writing outfile '%s' for writing (%s)\n", argv[0], Options.Outfile, strerror(errno));
			return 1;
		}
		for (i=0; i<NumberShareResultFiles; i++) {
			g=fopen(Shareresults[i], "r");
			if (g!=NULL) {
				while (fgets(Line, 1024, g)!=NULL) {
					if (strncmp(Line, "smb://", 6)==0) fprintf(f, "%s", Line);
				}
				fclose(g);
			}
			unlink(Shareresults[i]);
		}
		fclose(f);
	}

	return 0;
}
